/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares doubly-linked lists.
 */

#ifndef VRPSX_DIRECT_TESTS_TEST_DLIST_H
#define VRPSX_DIRECT_TESTS_TEST_DLIST_H 1

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* A link in a doubly-linked loop.
 */
typedef struct dlink_struct {
    struct dlink_struct *prev;
    struct dlink_struct *next;
} dlink_t;

/* Construct a link in a singleton doubly-linked loop.
 * No explicit destruction is required for such isolated links.
 */
extern void dlink_construct(
    dlink_t *link);

/* Removes the given link from its doubly-linked loop, shrinking
 * that loop and leaving the dlink_t instance in an invalid state.
 */
extern void dlink_destroy(
    dlink_t *link);

/* Swaps the "next" pointers of the given links, preserving invariants.
 *
 * If the two links were in different doubly-linked loops, then those
 * two loops are joined into one.  But if they were in the same loop,
 * then that loop is split into two, one containing each of the links.
 */
extern void dlink_splice(
    dlink_t *beforeA,
    dlink_t *beforeB);

/* Consider the sequence of links expressed by omitting the given
 * link from the doubly-linked loop that contains it.  This function
 * sorts that sequence using the given comparison callback, which
 * follows the strcmp convention regarding return values, and takes
 * as its first argument the third argument to this function.
 *
 * The comparison callback is never applied to the link that you pass
 * in (which is typically a special terminal link that is not part of
 * the semantic list).
 *
 * The comparison callback should not make use of the "prev" or "next"
 * members of the dlink_t instances it receives, because sorting may
 * violate dlink_t invariants and in any case changes link order.
 * Rather, the comparison callback should cast to pointers to structs
 * that contain dlink_t instances at the given addresses, and then
 * compare the other members of those structs.  Naturally, it is the
 * callers responsibility to ensure that all links in the loop may be
 * safely cast in this way--except for the special terminal link that
 * is the argument to this function, because it is never compared.
 */
extern void dlink_sort(
    dlink_t *term,
    int (*compare)(void *, dlink_t const *, dlink_t const *),
    void *callback_data);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_DIRECT_TESTS_TEST_DLIST_H */
